import os
import sys
import json
import uuid
import logging
import ipaddress
import traceback
import threading
import subprocess
from time import time
from time import sleep
from typing import List
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor

from .net_tools import Device
from .decorators import job_tracker
from .ip_parser import parse_ip_input
from .port_manager import PortManager

JOB_DIR = './lanscape-jobs/'
SAVE_FREQUENCY = 1 # seconds



class SubnetScanner:
    def __init__(
            self, 
            subnet: str, 
            port_list: str,
            parallelism: float = 1.0,
            uid: str = None
        ):
        self.subnet = parse_ip_input(subnet)
        self.port_list = port_list
        self.ports: list = PortManager().get_port_list(port_list).keys()
        self.running = False
        self.parallelism: float = float(parallelism)
        self.subnet_str = subnet
        self.job_stats = {'running': {}, 'finished': {}, 'timing': {}}
        self.uid = uid if uid else str(uuid.uuid4())
        self.results = ScannerResults(self)
        self.log: logging.Logger = logging.getLogger('SubnetScanner')
        self.log.debug(f'Instantiated with uid: {self.uid}')
        self.log.debug(f'Port Count: {len(self.ports)} | Device Count: {len(self.subnet)}')

    def scan_subnet_threaded(self) -> threading.Thread:
        """
        Start a new thread to scan the subnet.
        """
        t = threading.Thread(target=self.scan_subnet)
        t.start()
        return t
    
    @staticmethod
    def scan_subnet_standalone(subnet: str, port_list: str, parallelism: float = 1.0):
        """
        Use shell to start a new process for subnet scan.
        """
        scan = SubnetScanner(subnet, port_list, parallelism)
        scan.results.save()

        try:
            scanner_path = Path(__file__).parent.parent / 'scanner.py'
        except:
            scanner_path = 'scanner.py'

        try:
            subprocess.Popen(
                [sys.executable, scanner_path, scan.uid],
                stdout=None, stderr=None, stdin=None, close_fds=True
            )
            sleep(1)
            if not ScannerResults.get_scan(scan.uid)['running']:
                raise Exception('Could not start scanner in new process')

            
        except Exception:
            scan.log.debug(traceback.format_exc())
            scan.log.warning('unable to start standalone scanner')
            scan.scan_subnet_threaded()
        return scan.uid
    
    @staticmethod
    def instantiate_scan(scan_id: str) -> 'SubnetScanner':
        """
        Load a scan by its unique ID.
        """
        scan = SubnetScanner.get_scan(scan_id)
        return SubnetScanner(scan['subnet'], scan['port_list'], scan['parallelism'], scan['uid'])
    

    @staticmethod
    def get_scan(scan_id) -> dict:
        """
        Load a scan by its unique ID.
        """
        return ScannerResults.get_scan(scan_id)
    
    def scan_subnet(self):
        """
        Scan the subnet for devices and open ports.
        """
        self._set_stage('scanning devices')
        self.running = True
        self.results.auto_save()
        with ThreadPoolExecutor(max_workers=self._t_cnt(256)) as executor:
            futures = {executor.submit(self._get_host_details, str(ip)): str(ip) for ip in self.subnet}
            for future in futures:
                ip = futures[future]
                try:
                    future.result()
                except Exception as e:
                    self.log.error(f'[{ip}] scan failed. details below:\n{traceback.format_exc()}')
                    self.results.errors.append({
                        'basic': f"Error scanning IP {ip}: {e}",
                        'traceback': traceback.format_exc(),
                    })
                
        
        self._set_stage('testing ports')
        self._scan_network_ports()

        self._set_stage('complete')
        self.running = False
        self.results.save() # manual save to ensure latest is live
        return self.results
        

    def debug_active_scan(self):
        """
            Run this after running scan_subnet_threaded 
            to see the progress of the scan
        """
        while self.running:
            os.system('cls' if os.name == 'nt' else 'clear')
            print(f'{self.uid} - {self.subnet_str}')
            print(f"Scanned: {self.results.devices_scanned}/{self.results.devices_total}")
            print(f"Alive: {len(self.results.devices)}")
            print(f"Running jobs:  {self.job_stats['running']}")
            print(f"Finished jobs: {self.job_stats['finished']}")
            print(f"Job timing:    {self.job_stats['timing']}")
            sleep(1)


    
    def _get_host_details(self, host: str):
        """
        Get the MAC address and open ports of the given host.
        """
        device = Device(host)
        is_alive = self._ping(device)
        self.results.scanned()
        if not is_alive:
            return None
        self.log.debug(f'[{host}] is alive, getting metadata')
        
        device.get_metadata()
        self.results.devices.append(device)
        return True
        
    def _scan_network_ports(self):
        with ThreadPoolExecutor(max_workers=self._t_cnt(10)) as executor:
            futures = {executor.submit(self._scan_ports, device): device for device in self.results.devices}
            for future in futures:
                future.result()

    @job_tracker
    def _scan_ports(self, device: Device):
        self.log.debug(f'[{device.ip}] Initiating port scan')
        device.stage = 'scanning'
        with ThreadPoolExecutor(max_workers=self._t_cnt(128)) as executor:
            futures = {executor.submit(self._test_port, device, int(port)): port for port in self.ports}
            for future in futures:
                future.result()
        self.log.debug(f'[{device.ip}] Completed port scan')
        device.stage = 'complete'
    
    @job_tracker
    def _test_port(self,host: Device, port: int):
        """
        Test if a port is open on a given host.
        Device class handles tracking open ports.
        """
        return host.test_port(port)
    
        

    @job_tracker
    def _ping(self, host: Device):
        """
        Ping the given host and return True if it's reachable, False otherwise.
        """
        return host.is_alive(host.ip)
    
    def _t_cnt(self, base_threads: int) -> int:
        """
        Calculate the number of threads to use based on the base number 
        of threads and the parallelism factor.
        """
        return int(base_threads * self.parallelism)
    
    def _set_stage(self,stage):
        self.log.debug(f'[{self.uid}] Moving to Stage: {stage}')
        self.results.stage = stage
    
class ScannerResults:
    def __init__(self,scan: SubnetScanner):
        self.scan = scan
        self.port_list: str = scan.port_list
        self.subnet: str = scan.subnet_str
        self.parallelism: float = scan.parallelism
        self.uid = scan.uid

        self.devices_total: int = len(list(scan.subnet))
        self.devices_alive: int = 0
        self.devices_scanned: int = 0
        self.devices: List[Device] = []

        self.errors: List[str] = []        
        self.running: bool = False
        self.start_time: float = time()
        self.run_time: int = 0
        self.stage = 'instantiated'

        self.log = logging.getLogger('ScannerResults')
        self.log.debug(f'Instantiated Logger For Scan: {self.scan.uid}')

    @staticmethod
    def get_scan(scan_id: str):
        """
        load scan by scan id
        """

        for i in range(5):
            try:
                with open(f'{JOB_DIR}{scan_id}.json', 'r') as f:
                    return json.load(f)
            except json.JSONDecodeError:
                sleep(1)

        raise json.JSONDecodeError('Could not load scan data')

    def scanned(self):
        self.devices_scanned += 1
        
    def auto_save(self):
        threading.Thread(target=self._save_thread).start()
    
    def save(self):
        Path(JOB_DIR).mkdir(parents=True, exist_ok=True)
        self.log.debug(f"Saving Results. Threads: {self.scan.job_stats.get('running')}")

        self.running = self.scan.running
        self.run_time = int(round(time() - self.start_time,0))
        self.devices_alive = len(self.devices)

        out = vars(self).copy()
        out.pop('scan')
        out.pop('log')
        
        devices: Device = out.pop('devices')
        sortedDevices = sorted(devices, key=lambda obj: ipaddress.IPv4Address(obj.ip))
        out['devices'] = [vars(device).copy() for device in sortedDevices]
        for device in out['devices']: device.pop('log') 

        with open(f'{JOB_DIR}{self.uid}.json', 'w') as f:
            json.dump(out, f,indent=2)
    def _save_thread(self):
        while self.scan.running:
            self.save()
            sleep(SAVE_FREQUENCY)
        self.save()

            
def cleanup_old_jobs(older_than=0):
    """
    Removes removes jobs (scans) from the execution directory.
    Optional param to filter jobs older than (seconds).
    """
    for file in os.listdir(JOB_DIR):
        if file.endswith('.json'):
            file_path = f'{JOB_DIR}{file}'
            if time() - os.path.getmtime(file_path) > older_than:
                os.remove(file_path)


import requests
import pkg_resources
import logging 
import traceback

log = logging.getLogger('VersionManager')
PACKAGE='lanscape'

latest = None # used to 'remember' pypi version each runtime

def is_update_available(package=PACKAGE) -> bool:
    installed = get_installed_version(package)
    available = lookup_latest_version(package)
    if installed == '0.0.0': return False #local

    return installed != available

def lookup_latest_version(package=PACKAGE):
    # Fetch the latest version from PyPI
    global latest
    if not latest:
        url = f"https://pypi.org/pypi/{package}/json"
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raise an exception for HTTP errors
            latest = response.json()['info']['version']
            log.debug(f'Latest pypi version: {latest}')
        except:
            log.debug(traceback.format_exc())
            log.warning('Unable to fetch package version from PyPi')
    return latest

def get_installed_version(package=PACKAGE):
    installed_version = None
    try:
        installed_version = pkg_resources.get_distribution(package).version
    except:
        log.debug(traceback.format_exc())
        log.warning(f'Cannot find {package} installation')
        installed_version = '0.0.0'
    return installed_version




import unittest
import json
from ..app import app
from ..libraries.net_tools import get_network_subnet
from ..libraries.subnet_scan import cleanup_old_jobs
from ._helpers import right_size_subnet



class ApiTestCase(unittest.TestCase):
    app = app.test_client()

    def test_port_lifecycle(self):
        # Delete the new port list if it exists
        self.app.delete('/api/port/list/test_port_list_lifecycle')

        # Get the list of port lists
        response = self.app.get('/api/port/list')
        self.assertEqual(response.status_code, 200)
        port_list_start = json.loads(response.data)

        # Create a new port list
        new_port_list = {'80': 'http', '443': 'https'}
        response = self.app.post('/api/port/list/test_port_list_lifecycle', json=new_port_list)
        self.assertEqual(response.status_code, 200)


        # Get the list of port lists again
        response = self.app.get('/api/port/list')
        self.assertEqual(response.status_code, 200)
        port_list_new = json.loads(response.data)
        # Verify that the new port list is in the list of port lists
        self.assertEqual(len(port_list_new), len(port_list_start) + 1)

        # Get the new port list
        response = self.app.get('/api/port/list/test_port_list_lifecycle')
        self.assertEqual(response.status_code, 200)
        port_list = json.loads(response.data)
        self.assertEqual(port_list, new_port_list)

        # Update the new port list
        updated_port_list = {'22': 'ssh', '8080': 'http-alt'}
        response = self.app.put('/api/port/list/test_port_list_lifecycle', json=updated_port_list)
        self.assertEqual(response.status_code, 200)

        # Get the new port list again
        response = self.app.get('/api/port/list/test_port_list_lifecycle')
        self.assertEqual(response.status_code, 200)
        port_list = json.loads(response.data)

        # Verify that the new port list has been updated
        self.assertEqual(port_list, updated_port_list)

        # Delete the new port list
        response = self.app.delete('/api/port/list/test_port_list_lifecycle')
        self.assertEqual(response.status_code, 200)

    def test_scan(self):
        # Delete the new port list if it exists
        self.app.delete('/api/port/list/test_port_list_scan')

        # Create a new port list
        new_port_list = {'80': 'http', '443': 'https'}
        response = self.app.post('/api/port/list/test_port_list_scan', json=new_port_list)
        self.assertEqual(response.status_code, 200)

        # Create a new scan, wait for completion
        new_scan = {'subnet': right_size_subnet(get_network_subnet()), 'port_list': 'test_port_list_scan'}
        response = self.app.post('/api/scan/async', json=new_scan)
        self.assertEqual(response.status_code, 200)
        scan_info = json.loads(response.data)
        self.assertEqual(scan_info['status'], 'complete')
        scanid = scan_info['scan_id']
        self.assertIsNotNone(scanid)

        # Validate the scan worked without error
        response = self.app.get(f"/api/scan/{scanid}")
        self.assertEqual(response.status_code, 200)
        scan_data = json.loads(response.data)
        self.assertEqual(scan_data['errors'], [])
        self.assertEqual(scan_data['stage'],'complete')

        # Render UI
        uris = [
            f'/?scan_id={scanid}',
            f'/scan/{scanid}/overview',
            f'/scan/{scanid}/table'
        ]
        for uri in uris:
            response = self.app.get(uri)
            self.assertEqual(response.status_code,200)


        # Delete the new port list
        response = self.app.delete('/api/port/list/test_port_list_scan')
        self.assertEqual(response.status_code, 200)

        cleanup_old_jobs()
    
    def test_subnet_ports(self):
        """
        Test to ensure multi-subnet dectection is working
        """

        response = self.app.get('/api/tools/subnet/list')
        self.assertEqual(response.status_code, 200)

        subnets = json.loads(response.data)
        self.assertIsNot(len(subnets), 0)
        self.assertIsInstance(subnets[0],dict)
        subnet:dict = subnets[0]
        self.assertIsNotNone(subnet.get('address_cnt'))

    def test_subnet_validation(self):
        """
        test subnet validation and parsing is working as expected
        """
        subnet_tests = {
            # subnet : count (-1 == invalid)
            '10.0.0.0/24': 254,
            '10.0.0.2/24': 254,
            '10.0.0.1-100': 100,
            '192.168.1.1/25': 126,
            '10.0.0.1/24, 192.168.1.1-100': 354,
            '10.0.0.1/20': 4094,
            '10.0.0.1/19': 8190,
            '': -1, # blank
            '10.0.1/24': -1, # invalid
            '10.0.0.1/2': -1, # too big
            '10.0.0.1/19, 192.168.1.1/20': -1 # too big
        }

        for subnet, count in subnet_tests.items():
            uri = f'/api/tools/subnet/test?subnet={subnet}'
            response = self.app.get(uri)
            self.assertEqual(response.status_code, 200)

            data:dict = json.loads(response.data)
            self.assertEqual(data.get('count'),count)
            self.assertIsNotNone(data.get('msg'))
            if count == -1:
                self.assertFalse(data.get('valid'))
            




        
if __name__ == '__main__':
    unittest.main()



import threading
import webbrowser
import time
import logging
import traceback
import os
from .libraries.logger import configure_logging
from .libraries.runtime_args import parse_args, RuntimeArgs
# do this so any logs generated on import are displayed
args = parse_args()
configure_logging(args.loglevel, args.logfile)

from .libraries.version_manager import get_installed_version, is_update_available
from .webviewer import start_webview
from .app import start_webserver
from .libraries.subnet_scan import cleanup_old_jobs


log = logging.getLogger('core')
# determine if the execution is an instance of a flask reload
# happens on file change with reloader enabled
IS_FLASK_RELOAD = os.environ.get("WERKZEUG_RUN_MAIN")




def main():
    if not IS_FLASK_RELOAD:
        log.info(f'LANscape v{get_installed_version()}')
        try_check_update()
    else: 
        log.debug('Flask reloaded app.')
        
        
    try:
        if args.nogui:
            no_gui(args)
        else:
            start_webview(
                args
            )
        if not args.noclean:
            cleanup_old_jobs()
    except Exception:
        # showing error in debug only because this is handled gracefully
        log.debug('Failed to start webview client. Traceback below')
        log.debug(traceback.format_exc())
        if not args.nogui:
            log.error('Unable to start webview client. Try running with flag --nogui')


def try_check_update():
    try: 
        if is_update_available():
            log.info('An update is available!')
            log.info('Run "pip install --upgrade lanscape --no-cache" to supress this message.')
    except:
        log.debug(traceback.format_exc())
        log.warning('Unable to check for updates.')
    

def open_browser(url: str,wait=2):
    """
    Open a browser window to the specified
    url after waiting for the server to start
    """
    def do_open():
        try:
            time.sleep(wait)
            webbrowser.open(url, new=2)
        except:
            srv_url = f"0.0.0.0:{url.split(':')[1]}"
            log.debug(traceback.format_exc())
            log.info(f'unable to open web browser, server running on {srv_url}')

    threading.Thread(target=do_open).start()

def no_gui(args: RuntimeArgs):
    # determine if it was reloaded by flask debug reloader
    # if it was, dont open the browser again
    if not IS_FLASK_RELOAD:
        open_browser(f'http://127.0.0.1:{args.port}')
    start_webserver(
        args
    )


if __name__ == "__main__":
    main()
        



from time import time

def job_tracker(func):
        fxn = func.__name__
        def wrapper(*args, **kwargs):
            class_instance = args[0]
            running,finished,timing = init_job_tracker(class_instance)
            running[fxn] += 1
            start = time()
            result = func(*args, **kwargs)
            elapsed = time() - start
            running[fxn] -= 1
            finished[fxn] += 1
            timing[fxn] = round(((finished[fxn]-1 * timing[fxn] ) + elapsed ) / finished[fxn], 4)
            if running[fxn] == 0: running.pop(fxn)
            return result
        
        def init_job_tracker(class_instance):
            if not class_instance.job_stats:
                class_instance.job_stats = {'running': {}, 'finished': {}, 'timing': {}}
            running  = class_instance.job_stats.get('running', {})
            finished = class_instance.job_stats.get('finished', {})
            timing   = class_instance.job_stats.get('timing', {})

            running[fxn] = running.get(fxn, 0)
            finished[fxn] = finished.get(fxn, 0)
            timing[fxn] = timing.get(fxn, 0)

            return running, finished, timing
            
             
            
        return wrapper


# Used to spin up a new process outside of the main
# Flask application to perform the subnet scan.

from libraries.subnet_scan import SubnetScanner
import sys

if len(sys.argv) < 2:
    print('Usage: python scanner.py <scan_id>')
    sys.exit(1)

scanner = SubnetScanner.instantiate_scan(sys.argv[1])
scanner.scan_subnet()
